@@ -102,36 +102,12 @@ rc_getfile(const char *file, char **buffer, size_t *len)
 	return ret;
 }
 
-ssize_t
-rc_getline(char **line, size_t *len, FILE *fp)
-{
-	char *p;
-	size_t last = 0;
-
-	while (!feof(fp)) {
-		if (*line == NULL || last != 0) {
-			*len += BUFSIZ;
-			*line = xrealloc(*line, *len);
-		}
-		p = *line + last;
-		memset(p, 0, BUFSIZ);
-		if (fgets(p, BUFSIZ, fp) == NULL)
-			break;
-		last += strlen(p);
-		if (last && (*line)[last - 1] == '\n') {
-			(*line)[last - 1] = '\0';
-			break;
-		}
-	}
-	return last;
-}
-
 char *
 rc_proc_getent(const char *ent RC_UNUSED)
 {
 #ifdef __linux__
 	FILE *fp;
-	char *proc, *p, *value = NULL;
+	char *proc = NULL, *p, *value = NULL, *save;
 	size_t i, len;
 
 	if (!exists("/proc/cmdline"))
@@ -140,23 +116,20 @@ rc_proc_getent(const char *ent RC_UNUSED)
 	if (!(fp = fopen("/proc/cmdline", "r")))
 		return NULL;
 
-	proc = NULL;
 	i = 0;
-	if (rc_getline(&proc, &i, fp) == -1 || proc == NULL)
+	if (xgetline(&proc, &i, fp) == -1) {
+		free(proc);
 		return NULL;
+	}
+	save = proc;
 
-	if (proc != NULL) {
-		len = strlen(ent);
-
-		while ((p = strsep(&proc, " "))) {
-			if (strncmp(ent, p, len) == 0 && (p[len] == '\0' || p[len] == ' ' || p[len] == '=')) {
-				p += len;
-
-				if (*p == '=')
-					p++;
-
-				value = xstrdup(p);
-			}
+	len = strlen(ent);
+	while ((p = strsep(&save, " "))) {
+		if (strncmp(ent, p, len) == 0 && (p[len] == '\0' || p[len] == ' ' || p[len] == '=')) {
+			p += len;
+			if (*p == '=')
+				p++;
+			value = xstrdup(p);
 		}
 	}
 
@@ -173,7 +146,7 @@ rc_proc_getent(const char *ent RC_UNUSED)
 }
 
 RC_STRINGLIST *
-rc_config_list(const char *file)
+config_list(int dirfd, const char *pathname)
 {
 	FILE *fp;
 	char *buffer = NULL;
@@ -182,10 +155,10 @@ rc_config_list(const char *file)
 	char *token;
 	RC_STRINGLIST *list = rc_stringlist_new();
 
-	if (!(fp = fopen(file, "r")))
+	if (!(fp = do_fopenat(dirfd, pathname, O_RDONLY)))
 		return list;
 
-	while ((rc_getline(&buffer, &len, fp))) {
+	while (xgetline(&buffer, &len, fp) != -1) {
 		p = buffer;
 		/* Strip leading spaces/tabs */
 		while ((*p == ' ') || (*p == '\t'))
@@ -204,12 +177,18 @@ rc_config_list(const char *file)
 			}
 		}
 	}
-	fclose(fp);
 	free(buffer);
+	fclose(fp);
 
 	return list;
 }
 
+RC_STRINGLIST *
+rc_config_list(const char *file)
+{
+	return config_list(AT_FDCWD, file);
+}
+
 static void rc_config_set_value(RC_STRINGLIST *config, char *value)
 {
 	RC_STRING *cline;
@@ -287,6 +266,7 @@ static RC_STRINGLIST *rc_config_kcl(RC_STRINGLIST *config)
 	overrides = rc_stringlist_new();
 
 	/* A list of variables which may be overridden on the kernel command line */
+	rc_stringlist_add(overrides, "rc_interactive");
 	rc_stringlist_add(overrides, "rc_parallel");
 
 	TAILQ_FOREACH(override, overrides, entries) {
@@ -327,50 +307,52 @@ static RC_STRINGLIST *rc_config_kcl(RC_STRINGLIST *config)
 	return config;
 }
 
-static RC_STRINGLIST * rc_config_directory(RC_STRINGLIST *config)
+static void
+rc_config_directory(RC_STRINGLIST *config, int targetfd, const char *dir)
 {
 	DIR *dp;
 	struct dirent *d;
-	RC_STRINGLIST *rc_conf_d_files = rc_stringlist_new();
+	RC_STRINGLIST *rc_conf_d_files;
 	RC_STRING *fname;
 	RC_STRINGLIST *rc_conf_d_list;
-	char path[PATH_MAX];
 	RC_STRING *line;
 
-	if ((dp = opendir(RC_CONF_D)) != NULL) {
-		while ((d = readdir(dp)) != NULL) {
-			if (fnmatch("*.conf", d->d_name, FNM_PATHNAME) == 0) {
-				rc_stringlist_addu(rc_conf_d_files, d->d_name);
-			}
-		}
-		closedir(dp);
+	if (!(dp = do_opendirat(targetfd, dir)))
+		return;
 
-		if (rc_conf_d_files) {
-			rc_stringlist_sort(&rc_conf_d_files);
-			TAILQ_FOREACH(fname, rc_conf_d_files, entries) {
-				if (!fname->value)
-					continue;
-				sprintf(path, "%s/%s", RC_CONF_D, fname->value);
-				rc_conf_d_list = rc_config_list(path);
-				TAILQ_FOREACH(line, rc_conf_d_list, entries)
-					if (line->value)
-						rc_config_set_value(config, line->value);
-				rc_stringlist_free(rc_conf_d_list);
-			}
-			rc_stringlist_free(rc_conf_d_files);
+	rc_conf_d_files = rc_stringlist_new();
+	while ((d = readdir(dp)) != NULL) {
+		if (fnmatch("*.conf", d->d_name, FNM_PATHNAME) == 0) {
+			rc_stringlist_addu(rc_conf_d_files, d->d_name);
 		}
 	}
-	return config;
+
+	rc_stringlist_sort(&rc_conf_d_files);
+	TAILQ_FOREACH(fname, rc_conf_d_files, entries) {
+		if (!fname->value)
+			continue;
+
+		rc_conf_d_list = config_list(dirfd(dp), fname->value);
+		TAILQ_FOREACH(line, rc_conf_d_list, entries)
+			if (line->value)
+				rc_config_set_value(config, line->value);
+		rc_stringlist_free(rc_conf_d_list);
+	}
+
+	rc_stringlist_free(rc_conf_d_files);
+	closedir(dp);
+
+	return;
 }
 
-RC_STRINGLIST *
-rc_config_load(const char *file)
+static RC_STRINGLIST *
+config_load(int dirfd, const char *pathname)
 {
 	RC_STRINGLIST *list;
 	RC_STRINGLIST *config;
 	RC_STRING *line;
 
-	list = rc_config_list(file);
+	list = config_list(dirfd, pathname);
 	config = rc_stringlist_new();
 	TAILQ_FOREACH(line, list, entries) {
 		rc_config_set_value(config, line->value);
@@ -380,6 +362,12 @@ rc_config_load(const char *file)
 	return config;
 }
 
+RC_STRINGLIST *
+rc_config_load(const char *file)
+{
+	return config_load(AT_FDCWD, file);
+}
+
 char *
 rc_config_value(RC_STRINGLIST *list, const char *entry)
 {
@@ -408,35 +396,52 @@ _free_rc_conf(void)
 	rc_stringlist_free(rc_conf);
 }
 
+static void
+rc_conf_append(enum rc_dir dir)
+{
+	RC_STRINGLIST *conf = config_load(rc_dirfd(dir), "rc.conf");
+	TAILQ_CONCAT(rc_conf, conf, entries);
+	rc_stringlist_free(conf);
+}
+
 char *
 rc_conf_value(const char *setting)
 {
-	RC_STRINGLIST *old;
 	RC_STRING *s;
-	char *p;
 
-	if (!rc_conf) {
-		rc_conf = rc_config_load(RC_CONF);
-		atexit(_free_rc_conf);
+	if (rc_conf)
+		return rc_config_value(rc_conf, setting);
 
-		/* Support old configs. */
-		if (exists(RC_CONF_OLD)) {
-			old = rc_config_load(RC_CONF_OLD);
-			TAILQ_CONCAT(rc_conf, old, entries);
-			rc_stringlist_free(old);
-		}
+	rc_conf = rc_stringlist_new();
+	atexit(_free_rc_conf);
 
-		rc_conf = rc_config_directory(rc_conf);
-		rc_conf = rc_config_kcl(rc_conf);
+	/* Load user configurations first, as they should override
+	 * system wide configs. */
+	if (rc_is_user()) {
+		rc_conf_append(RC_DIR_USRCONF);
+		rc_config_directory(rc_conf, rc_dirfd(RC_DIR_USRCONF), "rc.conf.d");
+	}
 
-		/* Convert old uppercase to lowercase */
-		TAILQ_FOREACH(s, rc_conf, entries) {
-			p = s->value;
-			while (p && *p && *p != '=') {
-				if (isupper((unsigned char)*p))
-					*p = tolower((unsigned char)*p);
-				p++;
-			}
+	rc_conf_append(RC_DIR_SYSCONF);
+
+	/* Support old configs. */
+	if (exists(RC_CONF_OLD)) {
+		RC_STRINGLIST *old_conf = config_load(AT_FDCWD, RC_CONF_OLD);
+		TAILQ_CONCAT(rc_conf, old_conf, entries);
+		rc_stringlist_free(old_conf);
+	}
+
+	rc_config_directory(rc_conf, rc_dirfd(RC_DIR_SYSCONF), "rc.conf.d");
+
+	rc_conf = rc_config_kcl(rc_conf);
+
+	/* Convert old uppercase to lowercase */
+	TAILQ_FOREACH(s, rc_conf, entries) {
+		char *p = s->value;
+		while (p && *p && *p != '=') {
+			if (isupper((unsigned char)*p))
+				*p = tolower((unsigned char)*p);
+			p++;
 		}
 	}
 
