@@ -35,6 +35,7 @@
 #include <time.h>
 #include <unistd.h>
 #include <stdbool.h>
+#include <inttypes.h>
 
 #if defined(__linux__) || (defined(__FreeBSD_kernel__) && defined(__GLIBC__)) \
 	|| defined(__GNU__)
@@ -48,15 +49,14 @@
 #include "einfo.h"
 #include "queue.h"
 #include "rc.h"
+#include "rc_exec.h"
 #include "misc.h"
+#include "timeutils.h"
 #include "plugin.h"
 #include "selinux.h"
 #include "_usage.h"
 #include "helpers.h"
 
-#define PREFIX_LOCK	RC_SVCDIR "/prefix.lock"
-
-#define WAIT_INTERVAL	20000000	/* usecs to poll the lock file */
 #define WAIT_TIMEOUT	60		/* seconds until we timeout */
 #define WARN_TIMEOUT	10		/* warn about this every N seconds */
 
@@ -92,7 +92,8 @@ static RC_STRINGLIST *use_services;
 static RC_STRINGLIST *want_services;
 static RC_HOOK hook_out;
 static int exclusive_fd = -1, master_tty = -1;
-static bool sighup, in_background, deps, dry_run;
+static bool in_background, deps, dry_run;
+static volatile bool sighup, skip_mark, timedout;
 static pid_t service_pid;
 static int signal_pipe[2] = { -1, -1 };
 
@@ -117,12 +118,15 @@ handle_signal(int sig)
 		sighup = true;
 		break;
 
+	case SIGUSR1:
+		skip_mark = true;
+		break;
+
 	case SIGCHLD:
 		while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
 			if (signal_pipe[1] > -1 && pid == service_pid) {
 				if (write(signal_pipe[1], &status, sizeof(status)) == -1)
-					eerror("%s: send: %s",
-					    service, strerror(errno));
+					eerror("%s: send: %s", applet, strerror(errno));
 			}
 		}
 		break;
@@ -152,6 +156,10 @@ handle_signal(int sig)
 		exit(EXIT_FAILURE);
 		/* NOTREACHED */
 
+	case SIGALRM:
+		timedout = true;
+		break;
+
 	default:
 		eerror("%s: caught unknown signal %d", applet, sig);
 	}
@@ -163,19 +171,15 @@ handle_signal(int sig)
 static void
 unhotplug(void)
 {
-	char *file = NULL;
-
-	xasprintf(&file, RC_SVCDIR "/hotplugged/%s", applet);
-	if (exists(file) && unlink(file) != 0)
-		eerror("%s: unlink `%s': %s", applet, file, strerror(errno));
-	free(file);
+	if (unlinkat(rc_dirfd(RC_DIR_HOTPLUGGED), applet, 0) != 0 && errno != ENOENT)
+		eerror("%s: unlink '%s/hotplugged/%s': %s", applet, rc_svcdir(), applet, strerror(errno));
 }
 
 static void
 start_services(RC_STRINGLIST *list)
 {
 	RC_STRING *svc;
-	RC_SERVICE state = rc_service_state (service);
+	RC_SERVICE state = rc_service_state(applet);
 
 	if (!list)
 		return;
@@ -192,10 +196,8 @@ start_services(RC_STRINGLIST *list)
 			if (state & RC_SERVICE_INACTIVE ||
 			    state & RC_SERVICE_WASINACTIVE)
 			{
-				rc_service_schedule_start(service,
-				    svc->value);
-				ewarn("WARNING: %s will start when %s has started",
-				    svc->value, applet);
+				rc_service_schedule_start(applet, svc->value);
+				ewarn("WARNING: %s will start when %s has started", svc->value, applet);
 			} else
 				service_start(svc->value);
 		}
@@ -266,9 +268,9 @@ cleanup(void)
 	rc_stringlist_free(applet_list);
 	rc_stringlist_free(tmplist);
 	free(ibsave);
-	free(service);
 	free(prefix);
 	free(runlevel);
+	free(service);
 }
 
 /* Buffer and lock all output messages so that we get readable content */
@@ -291,7 +293,7 @@ write_prefix(const char *buffer, size_t bytes, bool *prefixed)
 	 * Lock the prefix.
 	 * open() may fail here when running as user, as RC_SVCDIR may not be writable.
 	 */
-	lock_fd = open(PREFIX_LOCK, O_WRONLY | O_CREAT, 0664);
+	lock_fd = openat(rc_dirfd(RC_DIR_SVCDIR), "prefix.lock", O_WRONLY | O_CREAT, 0664);
 
 	if (lock_fd != -1) {
 		while (flock(lock_fd, LOCK_EX) != 0) {
@@ -334,30 +336,46 @@ write_prefix(const char *buffer, size_t bytes, bool *prefixed)
 	return ret;
 }
 
+static int
+openrc_sh_exec(const char *openrc_sh, const char *arg1, const char *arg2)
+{
+	if (arg2)
+		einfov("Executing: %s %s %s %s %s", openrc_sh, openrc_sh, service, arg1, arg2);
+	else
+		einfov("Executing: %s %s %s %s", openrc_sh, openrc_sh, service, arg1);
+	execl(openrc_sh, openrc_sh, service, arg1, arg2, (char *) NULL);
+	eerror("%s: exec '%s': %s", service, openrc_sh, strerror(errno));
+	_exit(EXIT_FAILURE);
+}
+
 static int
 svc_exec(const char *arg1, const char *arg2)
 {
 	int ret, fdout = fileno(stdout);
 	struct termios tt;
 	struct winsize ws;
-	int i;
 	int flags = 0;
 	struct pollfd fd[2];
 	int s;
-	char *buffer;
+	char buffer[BUFSIZ];
 	size_t bytes;
 	bool prefixed = false;
 	int slave_tty;
 	sigset_t sigchldmask;
 	sigset_t oldmask;
+	int64_t wait_timeout, warn_timeout, now;
+	RC_STRINGLIST *keywords;
+	bool forever = false;
+
+	keywords = rc_deptree_depend(deptree, applet, "keyword");
+	if (rc_stringlist_find(keywords, "-timeout") ||
+			rc_stringlist_find(keywords, "notimeout"))
+		forever = true;
+	rc_stringlist_free(keywords);
 
 	/* Setup our signal pipe */
-	if (pipe(signal_pipe) == -1)
-		eerrorx("%s: pipe: %s", service, applet);
-	for (i = 0; i < 2; i++)
-		if ((flags = fcntl(signal_pipe[i], F_GETFD, 0) == -1 ||
-			fcntl(signal_pipe[i], F_SETFD, flags | FD_CLOEXEC) == -1))
-			eerrorx("%s: fcntl: %s", service, strerror(errno));
+	if (pipe2(signal_pipe, O_CLOEXEC) == -1)
+		eerrorx("%s: pipe2: %s", applet, applet);
 
 	/* Open a pty for our prefixed output
 	 * We do this instead of mapping pipes to stdout, stderr so that
@@ -381,95 +399,85 @@ svc_exec(const char *arg1, const char *arg2)
 			fcntl(slave_tty, F_SETFD, flags | FD_CLOEXEC);
 	}
 
+	now = tm_now();
+	wait_timeout = now + (WAIT_TIMEOUT * 1000);
+	warn_timeout = now + (WARN_TIMEOUT * 1000);
 	service_pid = fork();
 	if (service_pid == -1)
-		eerrorx("%s: fork: %s", service, strerror(errno));
+		eerrorx("%s: fork: %s", applet, strerror(errno));
 	if (service_pid == 0) {
+		char *openrc_sh;
+		xasprintf(&openrc_sh, "%s/openrc-run.sh", rc_svcdir());
+
 		if (slave_tty >= 0) {
 			dup2(slave_tty, STDOUT_FILENO);
 			dup2(slave_tty, STDERR_FILENO);
 		}
 
-		if (exists(RC_SVCDIR "/openrc-run.sh")) {
-			if (arg2)
-				einfov("Executing: %s %s %s %s %s",
-					RC_SVCDIR "/openrc-run.sh", RC_SVCDIR "/openrc-run.sh",
-					service, arg1, arg2);
-			else
-				einfov("Executing: %s %s %s %s",
-					RC_SVCDIR "/openrc-run.sh", RC_SVCDIR "/openrc-run.sh",
-					service, arg1);
-			execl(RC_SVCDIR "/openrc-run.sh",
-			    RC_SVCDIR "/openrc-run.sh",
-			    service, arg1, arg2, (char *) NULL);
-			eerror("%s: exec `" RC_SVCDIR "/openrc-run.sh': %s",
-			    service, strerror(errno));
-			_exit(EXIT_FAILURE);
-		} else {
-			if (arg2)
-				einfov("Executing: %s %s %s %s %s",
-					RC_LIBEXECDIR "/sh/openrc-run.sh",
-					RC_LIBEXECDIR "/sh/openrc-run.sh",
-			    	service, arg1, arg2);
-			else
-				einfov("Executing: %s %s %s %s",
-					RC_LIBEXECDIR "/sh/openrc-run.sh",
-					RC_LIBEXECDIR "/sh/openrc-run.sh",
-			    	service, arg1);
-			execl(RC_LIBEXECDIR "/sh/openrc-run.sh",
-			    RC_LIBEXECDIR "/sh/openrc-run.sh",
-			    service, arg1, arg2, (char *) NULL);
-			eerror("%s: exec `" RC_LIBEXECDIR "/sh/openrc-run.sh': %s",
-			    service, strerror(errno));
-			_exit(EXIT_FAILURE);
-		}
+		if (exists(openrc_sh))
+			openrc_sh_exec(openrc_sh, arg1, arg2);
+		else
+			openrc_sh_exec(RC_LIBEXECDIR "/sh/openrc-run.sh", arg1, arg2);
+
+		/* UNREACHABLE */
 	}
 
-	buffer = xmalloc(sizeof(char) * BUFSIZ);
 	fd[0].fd = signal_pipe[0];
+	fd[1].fd = master_tty;
 	fd[0].events = fd[1].events = POLLIN;
 	fd[0].revents = fd[1].revents = 0;
-	if (master_tty >= 0) {
-		fd[1].fd = master_tty;
-		fd[1].events = POLLIN;
-		fd[1].revents = 0;
-	}
 
 	for (;;) {
-		if ((s = poll(fd, master_tty >= 0 ? 2 : 1, -1)) == -1) {
+		int timeout;
+		if (forever) {
+			timeout = -1;
+		} else if ((timeout = warn_timeout - tm_now()) < 0) {
+			timeout = 0;
+		}
+
+		if (poll(fd, master_tty >= 0 ? 2 : 1, timeout) == -1) {
 			if (errno != EINTR) {
-				eerror("%s: poll: %s",
-				    service, strerror(errno));
+				eerror("%s: poll: %s", applet, strerror(errno));
 				ret = -1;
 				break;
 			}
+			continue;
 		}
 
-		if (s > 0) {
-			if (fd[1].revents & (POLLIN | POLLHUP)) {
-				bytes = read(master_tty, buffer, BUFSIZ);
-				write_prefix(buffer, bytes, &prefixed);
-			}
+		if (fd[1].revents & (POLLIN | POLLHUP)) {
+			bytes = read(master_tty, buffer, BUFSIZ);
+			write_prefix(buffer, bytes, &prefixed);
+		}
 
-			/* signal_pipe receives service_pid's exit status */
-			if (fd[0].revents & (POLLIN | POLLHUP)) {
-				if ((s = read(signal_pipe[0], &ret, sizeof(ret))) != sizeof(ret)) {
-					eerror("%s: receive failed: %s", service,
-						s < 0 ? strerror(errno) : "short read");
-					ret = -1;
-					break;
-				}
-				ret = WEXITSTATUS(ret);
-				if (ret != 0 && errno == ECHILD)
-					/* killall5 -9 could cause this */
-					ret = 0;
+		/* signal_pipe receives service_pid's exit status */
+		if (fd[0].revents & (POLLIN | POLLHUP)) {
+			if ((s = read(signal_pipe[0], &ret, sizeof(ret))) != sizeof(ret)) {
+				eerror("%s: receive failed: %s", applet, s < 0 ? strerror(errno) : "short read");
+				ret = -1;
 				break;
 			}
+			ret = WEXITSTATUS(ret);
+			if (ret != 0 && errno == ECHILD)
+				/* killall5 -9 could cause this */
+				ret = 0;
+			break;
+		}
+
+		if (forever)
+			continue;
+
+		if ((now = tm_now()) >= wait_timeout) {
+			kill(service_pid, SIGKILL);
+			ret = -1;
+			break;
+		} else if (now >= warn_timeout) {
+			/* the timer might get off by a few ms, add 500ms so we get round numbers matching svc_wait. */
+			ewarn("%s: waiting for service (%"PRId64" seconds)", applet, (wait_timeout - now + 500) / 1000);
+			if ((warn_timeout += (WARN_TIMEOUT * 1000)) > wait_timeout)
+				warn_timeout = wait_timeout;
 		}
 	}
 
-	free(buffer);
-
 	sigemptyset (&sigchldmask);
 	sigaddset (&sigchldmask, SIGCHLD);
 	sigprocmask (SIG_BLOCK, &sigchldmask, &oldmask);
@@ -495,67 +503,47 @@ svc_exec(const char *arg1, const char *arg2)
 static bool
 svc_wait(const char *svc)
 {
-	char *file = NULL;
 	int fd;
 	bool forever = false;
 	RC_STRINGLIST *keywords;
-	struct timespec interval, timeout, warn;
+	const char *base = basename_c(svc);
+	int timeout = WAIT_TIMEOUT;
+	bool retval = false;
 
 	/* Some services don't have a timeout, like fsck */
 	keywords = rc_deptree_depend(deptree, svc, "keyword");
-	if (rc_stringlist_find(keywords, "-timeout") ||
-	    rc_stringlist_find(keywords, "notimeout"))
+	if (rc_stringlist_find(keywords, "-timeout") || rc_stringlist_find(keywords, "notimeout"))
 		forever = true;
 	rc_stringlist_free(keywords);
 
-	xasprintf(&file, RC_SVCDIR "/exclusive/%s", basename_c(svc));
+	if ((fd = openat(rc_dirfd(RC_DIR_EXCLUSIVE), base, O_RDONLY | O_NONBLOCK)) == -1)
+		return errno == ENOENT;
 
-	interval.tv_sec = 0;
-	interval.tv_nsec = WAIT_INTERVAL;
-	timeout.tv_sec = WAIT_TIMEOUT;
-	timeout.tv_nsec = 0;
-	warn.tv_sec = WARN_TIMEOUT;
-	warn.tv_nsec = 0;
+	timedout = false;
 	for (;;) {
-		fd = open(file, O_RDONLY | O_NONBLOCK);
-		if (fd != -1) {
-			if (flock(fd, LOCK_SH | LOCK_NB) == 0) {
-				close(fd);
-				free(file);
-				return true;
-			}
-			close(fd);
+		if (!forever)
+			alarm(WARN_TIMEOUT);
+
+		if (flock(fd, LOCK_SH) == 0) {
+			retval = true;
+			break;
 		}
-		if (errno == ENOENT) {
-			free(file);
-			return true;
-		}
-		if (errno != EWOULDBLOCK) {
-			eerror("%s: open `%s': %s", applet, file,
-			    strerror(errno));
-			free(file);
-			exit(EXIT_FAILURE);
-		}
-		if (nanosleep(&interval, NULL) == -1) {
-			if (errno != EINTR)
-				goto finish;
-		}
-		if (!forever) {
-			timespecsub(&timeout, &interval, &timeout);
-			if (timeout.tv_sec <= 0)
-				goto finish;
-			timespecsub(&warn, &interval, &warn);
-			if (warn.tv_sec <= 0) {
-				ewarn("%s: waiting for %s (%d seconds)",
-				    applet, svc, (int)timeout.tv_sec);
-				warn.tv_sec = WARN_TIMEOUT;
-				warn.tv_nsec = 0;
-			}
+
+		if (errno != EINTR)
+			break;
+
+		if (!forever && timedout) {
+			timedout = false;
+			if ((timeout -= WARN_TIMEOUT) > 0)
+				ewarn("%s: waiting for %s (%d seconds)", applet, base, timeout);
+			else
+				break;
 		}
 	}
-finish:
-	free(file);
-	return false;
+
+	alarm(0);
+	close(fd);
+	return retval;
 }
 
 static void
@@ -608,13 +596,13 @@ svc_start_check(void)
 {
 	RC_SERVICE state;
 
-	state = rc_service_state(service);
+	state = rc_service_state(applet);
 
 	if (in_background) {
 		if (!(state & (RC_SERVICE_INACTIVE | RC_SERVICE_STOPPED)))
 			exit(EXIT_FAILURE);
 		if (rc_yesno(getenv("IN_HOTPLUG")))
-			rc_service_mark(service, RC_SERVICE_HOTPLUGGED);
+			rc_service_mark(applet, RC_SERVICE_HOTPLUGGED);
 		if (strcmp(runlevel, RC_LEVEL_SYSINIT) == 0)
 			ewarnx("WARNING: %s will be started in the"
 			    " next runlevel", applet);
@@ -623,8 +611,6 @@ svc_start_check(void)
 	if (exclusive_fd == -1)
 		exclusive_fd = svc_lock(applet, !deps);
 	if (exclusive_fd == -1) {
-		if (errno == EACCES)
-			eerrorx("%s: superuser access required", applet);
 		if (state & RC_SERVICE_STOPPING)
 			ewarnx("WARNING: %s is stopping", applet);
 		else
@@ -641,7 +627,7 @@ svc_start_check(void)
 		ewarnx("WARNING: %s has already started, but is inactive",
 		    applet);
 
-	rc_service_mark(service, RC_SERVICE_STARTING);
+	rc_service_mark(applet, RC_SERVICE_STARTING);
 	hook_out = RC_HOOK_SERVICE_START_OUT;
 	rc_plugin_run(RC_HOOK_SERVICE_START_IN, applet);
 }
@@ -653,8 +639,9 @@ svc_start_deps(void)
 	RC_STRING *svc, *svc2;
 	RC_SERVICE state;
 	int depoptions = RC_DEP_TRACE;
-	size_t len;
-	char *p, *tmp;
+	size_t len, svc_count;
+	char *tmp;
+	FILE *mem;
 	pid_t pid;
 
 	errno = 0;
@@ -757,33 +744,30 @@ svc_start_deps(void)
 	if (TAILQ_FIRST(tmplist)) {
 		/* Set the state now, then unlink our exclusive so that
 		   our scheduled list is preserved */
-		rc_service_mark(service, RC_SERVICE_STOPPED);
+		rc_service_mark(applet, RC_SERVICE_STOPPED);
 
 		rc_stringlist_free(use_services);
 		use_services = NULL;
-		len = 0;
+		mem = xopen_memstream(&tmp, &len);
+		svc_count = 0;
 		TAILQ_FOREACH(svc, tmplist, entries) {
-			rc_service_schedule_start(svc->value, service);
+			rc_service_schedule_start(svc->value, applet);
 			use_services = rc_deptree_depend(deptree,
 			    "iprovide", svc->value);
 			TAILQ_FOREACH(svc2, use_services, entries)
-			    rc_service_schedule_start(svc2->value, service);
+			    rc_service_schedule_start(svc2->value, applet);
 			rc_stringlist_free(use_services);
 			use_services = NULL;
-			len += strlen(svc->value) + 2;
-		}
-
-		len += 5;
-		tmp = p = xmalloc(sizeof(char) * len);
-		TAILQ_FOREACH(svc, tmplist, entries) {
-			if (p != tmp)
-				p += snprintf(p, len, ", ");
-			p += snprintf(p, len - (p - tmp),
-			    "%s", svc->value);
+			if (svc_count > 0)
+				fputs(", ", mem);
+			fputs(svc->value, mem);
+			++svc_count;
 		}
+		xclose_memstream(mem);
 		rc_stringlist_free(tmplist);
 		tmplist = NULL;
-		ewarnx("WARNING: %s will start when %s has started", applet, tmp);
+		ewarnx("WARNING: %s will start when %s %s started",
+			applet, tmp, svc_count > 1 ? "have" : "has");
 		free(tmp);
 	}
 
@@ -802,22 +786,24 @@ static void svc_start_real(void)
 		setenv("IN_BACKGROUND", ibsave, 1);
 	hook_out = RC_HOOK_SERVICE_START_DONE;
 	rc_plugin_run(RC_HOOK_SERVICE_START_NOW, applet);
+	skip_mark = false;
 	started = (svc_exec("start", NULL) == 0);
 	if (ibsave)
 		unsetenv("IN_BACKGROUND");
 
-	if (rc_service_state(service) & RC_SERVICE_INACTIVE)
+	if (rc_service_state(applet) & RC_SERVICE_INACTIVE)
 		ewarnx("WARNING: %s has started, but is inactive", applet);
 	else if (!started)
 		eerrorx("ERROR: %s failed to start", applet);
 
-	rc_service_mark(service, RC_SERVICE_STARTED);
+	if (!skip_mark)
+		rc_service_mark(applet, RC_SERVICE_STARTED);
 	exclusive_fd = svc_unlock(applet, exclusive_fd);
 	hook_out = RC_HOOK_SERVICE_START_OUT;
 	rc_plugin_run(RC_HOOK_SERVICE_START_DONE, applet);
 
 	/* Now start any scheduled services */
-	services = rc_services_scheduled(service);
+	services = rc_services_scheduled(applet);
 	TAILQ_FOREACH(svc, services, entries)
 	    if (rc_service_state(svc->value) & RC_SERVICE_STOPPED)
 		    service_start(svc->value);
@@ -862,7 +848,7 @@ svc_start(void)
 static int
 svc_stop_check(RC_SERVICE *state)
 {
-	*state = rc_service_state(service);
+	*state = rc_service_state(applet);
 
 	if (rc_runlevel_stopping() && *state & RC_SERVICE_FAILED)
 		exit(EXIT_FAILURE);
@@ -875,8 +861,6 @@ svc_stop_check(RC_SERVICE *state)
 	if (exclusive_fd == -1)
 		exclusive_fd = svc_lock(applet, !deps);
 	if (exclusive_fd == -1) {
-		if (errno == EACCES)
-			eerrorx("%s: superuser access required", applet);
 		if (*state & RC_SERVICE_STOPPING)
 			ewarnx("WARNING: %s is already stopping", applet);
 		eerrorx("ERROR: %s stopped by something else", applet);
@@ -889,14 +873,14 @@ svc_stop_check(RC_SERVICE *state)
 		return 1;
 	}
 
-	rc_service_mark(service, RC_SERVICE_STOPPING);
+	rc_service_mark(applet, RC_SERVICE_STOPPING);
 	hook_out = RC_HOOK_SERVICE_STOP_OUT;
 	rc_plugin_run(RC_HOOK_SERVICE_STOP_IN, applet);
 
 	if (!rc_runlevel_stopping()) {
-		if (rc_service_in_runlevel(service, RC_LEVEL_SYSINIT))
+		if (rc_service_in_runlevel(applet, RC_LEVEL_SYSINIT))
 			ewarn("WARNING: you are stopping a sysinit service");
-		else if (rc_service_in_runlevel(service, RC_LEVEL_BOOT))
+		else if (rc_service_in_runlevel(applet, RC_LEVEL_BOOT))
 			ewarn("WARNING: you are stopping a boot service");
 	}
 
@@ -966,14 +950,14 @@ svc_stop_deps(RC_SERVICE state)
 			continue;
 		if (rc_runlevel_stopping()) {
 			/* If shutting down, we should stop even
-			 * if a dependant failed */
+			 * if a dependent failed */
 			if (runlevel &&
 			    (strcmp(runlevel,
 				RC_LEVEL_SHUTDOWN) == 0 ||
 				strcmp(runlevel,
 				    RC_LEVEL_SINGLE) == 0))
 				continue;
-			rc_service_mark(service, RC_SERVICE_FAILED);
+			rc_service_mark(applet, RC_SERVICE_FAILED);
 		}
 		eerrorx("ERROR: cannot stop %s as %s "
 		    "is still up", applet, svc->value);
@@ -1008,6 +992,7 @@ svc_stop_real(void)
 		setenv("IN_BACKGROUND", ibsave, 1);
 	hook_out = RC_HOOK_SERVICE_STOP_DONE;
 	rc_plugin_run(RC_HOOK_SERVICE_STOP_NOW, applet);
+	skip_mark = false;
 	stopped = (svc_exec("stop", NULL) == 0);
 	if (ibsave)
 		unsetenv("IN_BACKGROUND");
@@ -1015,11 +1000,12 @@ svc_stop_real(void)
 	if (!stopped)
 		eerrorx("ERROR: %s failed to stop", applet);
 
-	if (in_background)
-		rc_service_mark(service, RC_SERVICE_INACTIVE);
-	else
-		rc_service_mark(service, RC_SERVICE_STOPPED);
-
+	if (!skip_mark) {
+		if (in_background)
+		    rc_service_mark(applet, RC_SERVICE_INACTIVE);
+		else
+		    rc_service_mark(applet, RC_SERVICE_STOPPED);
+	}
 	hook_out = RC_HOOK_SERVICE_STOP_OUT;
 	rc_plugin_run(RC_HOOK_SERVICE_STOP_DONE, applet);
 	hook_out = 0;
@@ -1050,7 +1036,7 @@ svc_stop(void)
 static void
 svc_restart(void)
 {
-	if (!(rc_service_state(service) & RC_SERVICE_STOPPED)) {
+	if (!(rc_service_state(applet) & RC_SERVICE_STOPPED)) {
 		get_started_services();
 		svc_stop();
 		if (dry_run)
@@ -1097,15 +1083,11 @@ service_plugable(void)
 int main(int argc, char **argv)
 {
 	bool doneone = false;
-	bool runscript = false;
 	int retval, opt, depoptions = RC_DEP_TRACE;
 	RC_STRING *svc;
-	char *path = NULL;
-	char *lnk = NULL;
-	char *dir, *save = NULL, *saveLnk = NULL;
+	const char *workingdir = "/";
 	char *pidstr = NULL;
 	size_t l = 0, ll;
-	const char *file;
 	struct stat stbuf;
 
 	/* Show help if insufficient args */
@@ -1115,8 +1097,6 @@ int main(int argc, char **argv)
 	}
 
 	applet = basename_c(argv[0]);
-	if (strcmp(applet, "runscript") == 0)
-		runscript = true;
 
 	if (stat(argv[1], &stbuf) != 0) {
 		fprintf(stderr, "openrc-run `%s': %s\n",
@@ -1124,59 +1104,71 @@ int main(int argc, char **argv)
 		exit(EXIT_FAILURE);
 	}
 
-	atexit(cleanup);
+	if (rc_yesno(getenv("RC_USER_SERVICES")))
+		rc_set_user();
 
-	/* We need to work out the real full path to our service.
-	 * This works fine, provided that we ONLY allow multiplexed services
-	 * to exist in the same directory as the master link.
-	 * Also, the master link as to be a real file in the init dir. */
-	path = realpath(argv[1], NULL);
-	if (!path) {
-		fprintf(stderr, "realpath: %s\n", strerror(errno));
-		exit(EXIT_FAILURE);
-	}
-	lnk = xmalloc(4096);
-	memset(lnk, 0, 4096);
-	if (readlink(argv[1], lnk, 4096-1)) {
-		dir = dirname(path);
-		if (strchr(lnk, '/')) {
-			save = xstrdup(dir);
-			saveLnk = xstrdup(lnk);
-			dir = dirname(saveLnk);
-			if (strcmp(dir, save) == 0)
-				file = basename_c(argv[1]);
-			else
-				file = basename_c(lnk);
-			dir = save;
-		} else
-			file = basename_c(argv[1]);
-		xasprintf(&service, "%s/%s", dir, file);
-		if (stat(service, &stbuf) != 0) {
-			free(service);
-			service = xstrdup(lnk);
-		}
-		free(save);
-		free(saveLnk);
-	}
-	free(lnk);
-	if (!service)
-		service = xstrdup(path);
-	applet = basename_c(service);
+	if (!(service = (realpath(argv[1], NULL))))
+		eerrorx("readpath: %s", strerror(errno));
+	applet = basename_c(argv[1]);
 
 	if (argc < 3)
 		usage(EXIT_FAILURE);
 
-	/* Change dir to / to ensure all init scripts don't use stuff in pwd */
-	if (chdir("/") == -1)
+	/* Ok, we are ready to go, so setup selinux if applicable */
+	selinux_setup(argv);
+
+	deps = true;
+
+	/* Punt the first arg as its our service name */
+	argc--;
+	argv++;
+
+	/* Right then, parse any options there may be */
+	while ((opt = getopt_long(argc, argv, getoptstring,
+		    longopts, (int *)0)) != -1)
+		switch (opt) {
+		case 'd':
+			setenv("RC_DEBUG", "YES", 1);
+			break;
+		case 'l':
+			exclusive_fd = atoi(optarg);
+			fcntl(exclusive_fd, F_SETFD,
+			    fcntl(exclusive_fd, F_GETFD, 0) | FD_CLOEXEC);
+			break;
+		case 's':
+			if (!(rc_service_state(applet) & RC_SERVICE_STARTED))
+				exit(EXIT_FAILURE);
+			break;
+		case 'S':
+			if (!(rc_service_state(applet) & RC_SERVICE_STOPPED))
+				exit(EXIT_FAILURE);
+			break;
+		case 'D':
+			deps = false;
+			break;
+		case 'Z':
+			dry_run = true;
+			break;
+		case_RC_COMMON_GETOPT
+		}
+
+	/* Change dir to / to ensure all init scripts don't use stuff in pwd
+	 * For user services, change to the user's HOME instead. */
+	if (rc_is_user() && !(workingdir = getenv("HOME")))
+		eerrorx("HOME is unset.");
+
+	if (chdir(workingdir) == -1)
 		eerror("chdir: %s", strerror(errno));
 
+	atexit(cleanup);
+
 	if ((runlevel = xstrdup(getenv("RC_RUNLEVEL"))) == NULL) {
 		env_filter();
 		env_config();
 		runlevel = rc_runlevel_get();
 	}
 
-	setenv("EINFO_LOG", service, 1);
+	setenv("EINFO_LOG", applet, 1);
 	setenv("RC_SVCNAME", applet, 1);
 
 	/* Set an env var so that we always know our pid regardless of any
@@ -1213,44 +1205,6 @@ int main(int argc, char **argv)
 		eprefix(prefix);
 	}
 
-	/* Ok, we are ready to go, so setup selinux if applicable */
-	selinux_setup(argv);
-
-	deps = true;
-
-	/* Punt the first arg as its our service name */
-	argc--;
-	argv++;
-
-	/* Right then, parse any options there may be */
-	while ((opt = getopt_long(argc, argv, getoptstring,
-		    longopts, (int *)0)) != -1)
-		switch (opt) {
-		case 'd':
-			setenv("RC_DEBUG", "YES", 1);
-			break;
-		case 'l':
-			exclusive_fd = atoi(optarg);
-			fcntl(exclusive_fd, F_SETFD,
-			    fcntl(exclusive_fd, F_GETFD, 0) | FD_CLOEXEC);
-			break;
-		case 's':
-			if (!(rc_service_state(service) & RC_SERVICE_STARTED))
-				exit(EXIT_FAILURE);
-			break;
-		case 'S':
-			if (!(rc_service_state(service) & RC_SERVICE_STOPPED))
-				exit(EXIT_FAILURE);
-			break;
-		case 'D':
-			deps = false;
-			break;
-		case 'Z':
-			dry_run = true;
-			break;
-		case_RC_COMMON_GETOPT
-		}
-
 	if (rc_yesno(getenv("RC_NODEPS")))
 		deps = false;
 
@@ -1278,6 +1232,8 @@ int main(int argc, char **argv)
 
 	/* Setup a signal handler */
 	signal_setup(SIGHUP, handle_signal);
+	signal_setup(SIGALRM, handle_signal);
+	signal_setup(SIGUSR1, handle_signal);
 	signal_setup(SIGINT, handle_signal);
 	signal_setup(SIGQUIT, handle_signal);
 	signal_setup(SIGTERM, handle_signal);
@@ -1289,9 +1245,6 @@ int main(int argc, char **argv)
 	applet_list = rc_stringlist_new();
 	rc_stringlist_add(applet_list, applet);
 
-	if (runscript)
-		ewarn("%s uses runscript, please convert to openrc-run.", service);
-
 	/* Now run each option */
 	retval = EXIT_SUCCESS;
 	while (optind < argc) {
@@ -1314,7 +1267,7 @@ int main(int argc, char **argv)
 		    strcmp(optarg, "help") == 0 ||
 		    strcmp(optarg, "depend") == 0)
 		{
-			save = prefix;
+			char *save = prefix;
 			eprefix(NULL);
 			prefix = NULL;
 			svc_exec(optarg, NULL);
@@ -1352,7 +1305,6 @@ int main(int argc, char **argv)
 			rc_stringlist_free(services);
 			services = NULL;
 		} else if (strcmp (optarg, "status") == 0) {
-			save = prefix;
 			eprefix(NULL);
 			prefix = NULL;
 			retval = svc_exec("status", NULL);
@@ -1360,8 +1312,7 @@ int main(int argc, char **argv)
 			if (strcmp(optarg, "conditionalrestart") == 0 ||
 			    strcmp(optarg, "condrestart") == 0)
 			{
-				if (rc_service_state(service) &
-				    RC_SERVICE_STARTED)
+				if (rc_service_state(applet) & RC_SERVICE_STARTED)
 					svc_restart();
 			} else if (strcmp(optarg, "restart") == 0) {
 				svc_restart();
@@ -1377,22 +1328,13 @@ int main(int argc, char **argv)
 				if (svc_stop() == 1)
 					continue; /* Service has been stopped already */
 				if (deps) {
-					if (!in_background &&
-					    !rc_runlevel_stopping() &&
-					    rc_service_state(service) &
-					    RC_SERVICE_STOPPED)
+					if (!in_background && !rc_runlevel_stopping() && rc_service_state(applet) & RC_SERVICE_STOPPED)
 						unhotplug();
 
-					if (in_background &&
-					    rc_service_state(service) &
-					    RC_SERVICE_INACTIVE)
-					{
-						TAILQ_FOREACH(svc,
-						    restart_services,
-						    entries)
-						    if (rc_service_state(svc->value) &
-							RC_SERVICE_STOPPED)
-							    rc_service_schedule_start(service, svc->value);
+					if (in_background && rc_service_state(applet) & RC_SERVICE_INACTIVE) {
+						TAILQ_FOREACH(svc, restart_services, entries)
+							if (rc_service_state(svc->value) & RC_SERVICE_STOPPED)
+								rc_service_schedule_start(applet, svc->value);
 					}
 				}
 			} else if (strcmp(optarg, "zap") == 0) {
