@@ -2,7 +2,7 @@
   start-stop-daemon
  * Starts, stops, tests and signals daemons
  *
- * This is essentially a ground up re-write of Debians
+ * This is essentially a ground up re-write of Debian's
  * start-stop-daemon for cleaner code and to integrate into our RC
  * system so we can monitor daemons a little.
  */
@@ -21,7 +21,7 @@
 
 #define ONE_MS           1000000
 
-#ifdef __linux__
+#if defined(__linux__) && !defined(_GNU_SOURCE)
 /* For extra SCHED_* defines. */
 # define _GNU_SOURCE
 #endif
@@ -31,6 +31,7 @@
 #include <getopt.h>
 #include <limits.h>
 #include <grp.h>
+#include <pthread.h>
 #include <pwd.h>
 #include <sched.h>
 #include <signal.h>
@@ -39,6 +40,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <syslog.h>
 #include <sys/ioctl.h>
 #include <sys/resource.h>
 #include <sys/stat.h>
@@ -60,7 +62,7 @@
 static struct pam_conv conv = { NULL, NULL};
 #endif
 
-#ifdef HAVE_CAP
+#ifdef __linux__
 # include <sys/capability.h>
 #endif
 
@@ -68,7 +70,7 @@ static struct pam_conv conv = { NULL, NULL};
 #include "queue.h"
 #include "rc.h"
 #include "misc.h"
-#include "pipes.h"
+#include "rc_exec.h"
 #include "schedules.h"
 #include "_usage.h"
 #include "helpers.h"
@@ -87,11 +89,12 @@ enum {
   LONGOPT_SCHEDULER,
   LONGOPT_SCHEDULER_PRIO,
   LONGOPT_SECBITS,
+  LONGOPT_NOTIFY,
 };
 
 const char *applet = NULL;
 const char *extraopts = NULL;
-const char getoptstring[] = "I:KN:PR:Sa:bc:d:e:g:ik:mn:op:s:tu:r:w:x:1:2:3:4:" \
+const char getoptstring[] = "I:KN:PR:Sa:bc:d:e:g:ik:mn:op:s:tu:r:w:x:0:1:2:3:4:" \
 	getoptstring_COMMON;
 const struct option longopts[] = {
 	{ "capabilities", 1, NULL, LONGOPT_CAPABILITIES},
@@ -121,6 +124,7 @@ const struct option longopts[] = {
 	{ "chroot",       1, NULL, 'r'},
 	{ "wait",         1, NULL, 'w'},
 	{ "exec",         1, NULL, 'x'},
+	{ "stdin",        1, NULL, '0'},
 	{ "stdout",       1, NULL, '1'},
 	{ "stderr",       1, NULL, '2'},
 	{ "stdout-logger",1, NULL, '3'},
@@ -128,6 +132,7 @@ const struct option longopts[] = {
 	{ "progress",     0, NULL, 'P'},
 	{ "scheduler",    1, NULL, LONGOPT_SCHEDULER},
 	{ "scheduler-priority",    1, NULL, LONGOPT_SCHEDULER_PRIO},
+	{ "notify",        1, NULL, LONGOPT_NOTIFY},
 	longopts_COMMON
 };
 const char * const longopts_help[] = {
@@ -158,6 +163,7 @@ const char * const longopts_help[] = {
 	"Chroot to this directory",
 	"Milliseconds to wait for daemon start",
 	"Binary to start/stop",
+	"Redirect stdin to file",
 	"Redirect stdout to file",
 	"Redirect stderr to file",
 	"Redirect stdout to process",
@@ -314,6 +320,7 @@ int main(int argc, char **argv)
 	gid_t gid = 0;
 	char *home = NULL;
 	int tid = 0;
+	char *redirect_stdin = NULL;
 	char *redirect_stderr = NULL;
 	char *redirect_stdout = NULL;
 	char *stderr_process = NULL;
@@ -341,7 +348,7 @@ int main(int argc, char **argv)
 	unsigned int start_wait = 0;
 	const char *scheduler = NULL;
 	int sched_prio = -1;
-#ifdef HAVE_CAP
+#ifdef __linux__
 	cap_iab_t cap_iab = NULL;
 	unsigned secbits = 0;
 #endif
@@ -351,6 +358,8 @@ int main(int argc, char **argv)
 	int pipefd[2];
 	char readbuf[1];
 	ssize_t ss;
+	struct notify notify = {0};
+	int ret = EXIT_SUCCESS;
 
 	applet = basename_c(argv[0]);
 	atexit(cleanup);
@@ -359,6 +368,11 @@ int main(int argc, char **argv)
 	signal_setup(SIGQUIT, handle_signal);
 	signal_setup(SIGTERM, handle_signal);
 
+	if (rc_yesno(getenv("RC_USER_SERVICES")))
+		rc_set_user();
+
+	openlog(applet, LOG_PID, LOG_DAEMON);
+
 	if ((tmp = getenv("SSD_NICELEVEL")))
 		if (sscanf(tmp, "%d", &nicelevel) != 1)
 			eerror("%s: invalid nice level `%s' (SSD_NICELEVEL)",
@@ -398,7 +412,7 @@ int main(int argc, char **argv)
 		    (int *) 0)) != -1)
 		switch (opt) {
 		case LONGOPT_CAPABILITIES:
-#ifdef HAVE_CAP
+#ifdef __linux__
 			cap_iab = cap_iab_from_text(optarg);
 			if (cap_iab == NULL)
 				eerrorx("Could not parse iab: %s", strerror(errno));
@@ -408,7 +422,7 @@ int main(int argc, char **argv)
 			break;
 
 		case LONGOPT_SECBITS:
-#ifdef HAVE_CAP
+#ifdef __linux__
 			if (*optarg == '\0')
 				eerrorx("Secbits are empty");
 
@@ -590,6 +604,10 @@ int main(int argc, char **argv)
 			exec = optarg;
 			break;
 
+		case '0':   /* --stdin /path/to/stdin.input-file */
+			redirect_stdin = optarg;
+			break;
+
 		case '1':   /* --stdout /path/to/stdout.lgfile */
 			redirect_stdout = optarg;
 			break;
@@ -614,6 +632,10 @@ int main(int argc, char **argv)
 			sscanf(optarg, "%d", &sched_prio);
 			break;
 
+		case LONGOPT_NOTIFY:
+			notify = notify_parse(svcname ? svcname : applet, optarg);
+			break;
+
 		case_RC_COMMON_GETOPT
 		}
 
@@ -741,7 +763,7 @@ int main(int argc, char **argv)
 		fp = fopen(exec_file, "r");
 		if (fp) {
 			line = NULL;
-			if (getline(&line, &size, fp) == -1)
+			if (xgetline(&line, &size, fp) == -1)
 				eerrorx("%s: %s", applet, strerror(errno));
 			p = line;
 			fclose(fp);
@@ -750,10 +772,6 @@ int main(int argc, char **argv)
 				/* Strip leading spaces */
 				while (*p == ' ' || *p == '\t')
 					p++;
-				/* Remove the trailing newline */
-				len = strlen(p) - 1;
-				if (p[len] == '\n')
-					p[len] = '\0';
 				token = strsep(&p, " ");
 				free(exec_file);
 				xasprintf(&exec_file, "%s", token);
@@ -951,7 +969,7 @@ int main(int argc, char **argv)
 		if (changeuser && initgroups(changeuser, gid))
 			eerrorx("%s: initgroups (%s, %d)",
 			    applet, changeuser, gid);
-#ifdef HAVE_CAP
+#ifdef __linux__
 		if (uid && cap_setuid(uid))
 #else
 		if (uid && setuid(uid))
@@ -962,7 +980,7 @@ int main(int argc, char **argv)
 		/* Close any fd's to the passwd database */
 		endpwent();
 
-#ifdef HAVE_CAP
+#ifdef __linux__
 		if (cap_iab != NULL) {
 			i = cap_iab_set_proc(cap_iab);
 
@@ -1015,17 +1033,9 @@ int main(int argc, char **argv)
 #endif
 
 		TAILQ_FOREACH(env, env_list, entries) {
-			if ((strncmp(env->value, "RC_", 3) == 0 &&
-				strncmp(env->value, "RC_SERVICE=", 11) != 0 &&
-				strncmp(env->value, "RC_SVCNAME=", 11) != 0) ||
-				strncmp(env->value, "SSD_NICELEVEL=", 14) == 0 ||
-				strncmp(env->value, "SSD_IONICELEVEL=", 16) == 0 ||
-				strncmp(env->value, "SSD_OOM_SCORE_ADJ=", 18) == 0)
-			{
-				p = strchr(env->value, '=');
-				*p = '\0';
+			if (strncmp(env->value, "RC_", 3) == 0 && strncmp(env->value, "SSD_", 4) == 0) {
+				*strchr(env->value, '=') = '\0';
 				unsetenv(env->value);
-				continue;
 			}
 		}
 		rc_stringlist_free(env_list);
@@ -1060,6 +1070,13 @@ int main(int argc, char **argv)
 		stdin_fd = devnull_fd;
 		stdout_fd = devnull_fd;
 		stderr_fd = devnull_fd;
+		if (redirect_stdin) {
+			if ((stdin_fd = open(redirect_stdin,
+				    O_RDONLY)) == -1)
+				eerrorx("%s: unable to open the input file"
+				    " for stdin `%s': %s",
+				    applet, redirect_stdin, strerror(errno));
+		}
 		if (redirect_stdout) {
 			if ((stdout_fd = open(redirect_stdout,
 				    O_WRONLY | O_CREAT | O_APPEND,
@@ -1068,7 +1085,7 @@ int main(int argc, char **argv)
 				    " for stdout `%s': %s",
 				    applet, redirect_stdout, strerror(errno));
 		}else if (stdout_process) {
-			stdout_fd = rc_pipe_command(stdout_process);
+			stdout_fd = rc_pipe_command(stdout_process, devnull_fd);
 			if (stdout_fd == -1)
 				eerrorx("%s: unable to open the logging process"
 				    " for stdout `%s': %s",
@@ -1082,7 +1099,7 @@ int main(int argc, char **argv)
 				    " for stderr `%s': %s",
 				    applet, redirect_stderr, strerror(errno));
 		}else if (stderr_process) {
-			stderr_fd = rc_pipe_command(stderr_process);
+			stderr_fd = rc_pipe_command(stderr_process, devnull_fd);
 			if (stderr_fd == -1)
 				eerrorx("%s: unable to open the logging process"
 				    " for stderr `%s': %s",
@@ -1100,6 +1117,11 @@ int main(int argc, char **argv)
 
 		cloexec_fds_from(3);
 
+		if (notify.type == NOTIFY_FD) {
+			if (close(notify.pipe[0]) == -1 || dup2(notify.pipe[1], notify.fd) == -1)
+				eerrorx("Failed to initialize notify fd.");
+		}
+
 		if (scheduler != NULL) {
 			int scheduler_index;
 			struct sched_param sched =  {.sched_priority = sched_prio};
@@ -1123,7 +1145,7 @@ int main(int argc, char **argv)
 			if (sched_prio == -1)
 				sched.sched_priority = sched_get_priority_min(scheduler_index);
 
-			if (sched_setscheduler(mypid, scheduler_index, &sched))
+			if (pthread_setschedparam(pthread_self(), scheduler_index, &sched))
 				eerrorx("Failed to set scheduler: %s", strerror(errno));
 		} else if (sched_prio != -1) {
 			const struct sched_param sched =  {.sched_priority = sched_prio};
@@ -1184,7 +1206,10 @@ int main(int argc, char **argv)
 			start_wait = 0;
 	}
 
-	if (start_wait > 0) {
+	if (notify.type != NOTIFY_NONE) {
+		if (!notify_wait(applet, notify))
+			ret = EXIT_FAILURE;
+	} else if (start_wait > 0) {
 		struct timespec ts;
 		bool alive = false;
 
@@ -1224,6 +1249,6 @@ int main(int argc, char **argv)
 		rc_service_daemon_set(svcname, exec,
 		    (const char *const *)margv, pidfile, true);
 
-	exit(EXIT_SUCCESS);
+	exit(ret);
 	/* NOTREACHED */
 }
